﻿using CommandLine;
using Janus.Core.Enums;
using Janus.Core.Exporters;
using Janus.Core.Importers;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using Janus.Core.Utils;
using System;
using System.Collections.Generic;
using System.IO;

namespace Janus.Analyzers.Counter
{
    class Program
    {
        /// <summary>
        /// Command line arguments parsed as options
        /// </summary>
        private static CLIOptions _options;

        /// <summary>
        /// Main entry pojnt of the program
        /// </summary>
        /// <param name="args"></param>
        static void Main(string[] args)
        {
            var result = CommandLine.Parser.Default.ParseArguments<CLIOptions>(args)
                                                            .WithParsed(RunOptions)
                                                            .WithNotParsed(HandleParseError);
        }

        /// <summary>
        /// Main logic, if argument parsing is without an error
        /// </summary>
        /// <param name="opts">CLI Arguments as options</param>
        /// <returns></returns>
        static void RunOptions(CLIOptions opts)
        {
            ImportResult<DataSerializationWrapper> importResult;
            _options = opts; //Pass args to the rest of the program

            if(opts.OutputType != EOutputType.Text)
            {
                //Unsupported output
                Console.Error.WriteLine($"Unsupported option to print, \"{opts.OutputType}\" not available!");
                return;
            }

            if (opts.InputFile != null)
            {
                //Try to read from file

                if (!File.Exists(opts.InputFile))
                {
                    Console.Error.WriteLine($"ERROR: File does not exist! (Path: {opts.InputFile})");
                    return;
                }

                //File exists, try to read

                importResult = JsonImporter.Instance.ImportWrapperFromFile(opts.InputFile);
            }
            else
            {
                //File input not specified, fallback to std in
                importResult = JsonImporter.Instance.ImportWrapperFromStream(Console.In);
            }
            //File loaded
            if (!importResult.OperationSuccessful)
            {
                //Import problem .... end execution
                foreach (var err in importResult.Errors)
                {
                    Console.Error.WriteLine(err);
                }
                return;
            }

            //OK Ready to process
            CounterAnalyzer cAnalyzer = new CounterAnalyzer();

            AnalysisResultModel resultOfAnalyzer = null;
            try
            {
                //Sanity check -> this will prevent from failing
                cAnalyzer.ProcessDataWrapper(importResult.Result);
                resultOfAnalyzer = cAnalyzer.AnalyzerResult;
            }
            catch (Exception e)
            {
                Console.Error.WriteLine($"ERROR: During analysis ...");
                Console.Error.WriteLine($"Detailed error: {e.Message}");
                return;
            }

            try
            {
                //Data processed -> Send to output
                switch (opts.OutputType)
                {
                    case EOutputType.Text:
                        PrintResultText(resultOfAnalyzer, opts.OutputFile);
                        break;
                    default:
                        Console.Error.WriteLine($"Unsupported option to print, \"{opts.OutputType}\" not available!");
                        break;
                }        
            }
            catch (Exception e)
            {
                Console.Error.WriteLine($"ERROR: During data print ...");
                Console.Error.WriteLine($"Detailed error: {e.Message}");

            }
        }

        /// <summary>
        /// This method is called in case of argument parsing failure
        /// </summary>
        /// <param name="errs">Parsing errors as text</param>
        static void HandleParseError(IEnumerable<Error> errs)
        {
            Console.Error.WriteLine("ERROR During arguments matching ...");
            foreach (var item in errs)
            {
                Console.Error.WriteLine(item);
            }
        }

        /// <summary>
        /// Prints textual results to the file, if file is not speciefied fallback to the std out
        /// </summary>
        /// <param name="results">Result to be printed</param>
        /// <param name="outFile">File path to be printed result to</param>
        static void PrintResultText(AnalysisResultModel results, string outFile = null)
        {
            // Try to save file, null means no file saving is required
            if (outFile != null)
            {
                try
                {
                    File.WriteAllText(outFile, results.ToString());
                }
                catch (Exception e)
                {
                    Console.Error.WriteLine("Export failed!");
                    Console.WriteLine(e);
                }
            }
            else
            {
                Console.WriteLine(results.ToString());
            }
        }
    }
}
