﻿using Janus.Core.Importers;
using Janus.Core.Models;
using Janus.Core.Models.Results;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace Janus.Analyzers.KeyValue
{
    public class KeyValueAnalyzer
    {
        /// <summary>
        /// Patterns to be used to search in the dpeendencies
        /// Format: Key=Value ("License=MIT")
        /// </summary>
        IList<Tuple<string, string>> _searchPatterns = new List<Tuple<string,string>>();

        /// <summary>
        /// Dictionary of objects used to search in and it also contains the Dictionary of search pattern and its match as Value
        /// Key:    DependencyItem, one item analyzed
        /// Value:  HashSet
        ///                  - Key:     SearchPattern (Tuple)
        /// </summary>
        Dictionary<DependencyItem, HashSet<Tuple<string, string>>> markedNodes = new Dictionary<DependencyItem, HashSet<Tuple<string, string>>>();

        ////BFS fields
        
        Queue<DependencyItem> bfsToBeAnalysed = new Queue<DependencyItem>();

        //Key:   Visited dependency refference
        HashSet<DependencyItem> bfsVisitedDependencies = new HashSet<DependencyItem>();


        public KeyValueAnalyzer(IList<Tuple<string, string>> searchPatterns)
        {
            _searchPatterns = searchPatterns;
        }

        /// <summary>
        /// Processes the second part of deserialization and redirects the analysis to the data specific methods
        /// </summary>
        /// <param name="dataWrapper">Data wrapper to be redirected</param>
        public void ProcessDataWrapper(DataSerializationWrapper dataWrapper)
        {
            if (dataWrapper.DataType == null)
            {
                //NULL!!!!
                Console.Error.WriteLine($"ERROR: Datatype in the wrapper is null, cannot deserialize!");
                return;
            }
            else if (dataWrapper.DataType == new List<DependencyItem>().GetType().ToString())
            {
                ImportResult<IList<DependencyItem>> importedData = JsonImporter.Instance.ImportData<IList<DependencyItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeDependencies(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else if (dataWrapper.DataType == new List<ProjectItem>().GetType().ToString())
            {
                ImportResult<IList<ProjectItem>> importedData = JsonImporter.Instance.ImportData<IList<ProjectItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeProjects(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else if (dataWrapper.DataType == new List<SolutionItem>().GetType().ToString())
            {
                ImportResult<IList<SolutionItem>> importedData = JsonImporter.Instance.ImportData<IList<SolutionItem>>(dataWrapper);
                if (importedData.OperationSuccessful)
                {
                    dataWrapper.Data = importedData.Result;
                    AnalyzeSolutions(importedData.Result);
                }
                else
                {
                    //Import problem .... end execution
                    foreach (var err in importedData.Errors)
                    {
                        Console.Error.WriteLine(err);
                    }
                    return;
                }
            }
            else
            {
                //Unknown ...
                Console.Error.WriteLine($"ERROR: Datatype in the wrapper is unknown, cannot deserialize!");
                return;
            }

            //Should be imported fully now ... start the analysis
            StartDependencyAnalysis();
        }

        
        /// <summary>
        /// Performs the analysis queue for each solution of the given list of the solutions
        /// Internally calls <see cref="AnalyzeSolution"/>
        /// </summary>
        /// <param name="solutions">Solutions to analyze</param>
        private void AnalyzeSolutions(IList<SolutionItem> solutions)
        {
            foreach (var solution in solutions)
            {
                AnalyzeSolution(solution);
            }
        }

        /// <summary>
        /// Performs the analysis queue for each project in the given solution
        /// Internally calls <see cref="AnalyzeProject"/>
        /// </summary>
        /// <param name="solution">Solution to analyze</param>
        private void AnalyzeSolution(SolutionItem solution)
        {
            foreach (var projectEdge in solution.Projects)
            {
                if(projectEdge.Link != null)
                {
                    AnalyzeProject(projectEdge.Link);
                }
            }
        }

        /// <summary>
        /// Performs the analysis queue for each project of the given list of the projects
        /// Internally calls <see cref="AnalyzeProject"/>
        /// </summary>
        /// <param name="projects">Projects to analyze</param>
        private void AnalyzeProjects(IList<ProjectItem> projects)
        {
            foreach (var project in projects)
            {
                AnalyzeProject(project);
            }
        }

        /// <summary>
        /// Enques the direct dependencies of the project to the work queue
        /// </summary>
        /// <param name="project">Projects to analyze</param>
        private void AnalyzeProject(ProjectItem project)
        {
            foreach (var dependencyEdge in project.Dependencies)
            {
                if(dependencyEdge.Link != null)
                {
                    bfsToBeAnalysed.Enqueue(dependencyEdge.Link); //Add to the work list
                }
            }
        }

        /// <summary>
        /// Enques each dependency to the work queue
        /// </summary>
        /// <param name="dependencies">Dependencies to analyze</param>
        private void AnalyzeDependencies(IList<DependencyItem> dependencies)
        {
            foreach (var dependency in dependencies)
            {
                bfsToBeAnalysed.Enqueue(dependency); //Add to the work list
                //AnalyzeDependency(dependency); 
            }
        }

        /// <summary>
        /// Performs the dependency analysis with the filled work queue of dpeendencies 
        /// </summary>
        private void StartDependencyAnalysis()
        {
            while (bfsToBeAnalysed.Any())
            {
                DependencyItem current = bfsToBeAnalysed.Dequeue();

                if (bfsVisitedDependencies.Contains(current))
                {
                    //Already visited!!!
                    //Console.WriteLine("Been there, done that!");
                    continue;
                }

                bfsVisitedDependencies.Add(current); //Mark as visited

                //Process
                AnalyzeDependency(current);

                //Add subdependencies to the work queue
                foreach (var subDependencyEdge in current.Dependencies)
                {
                    DependencyItem subDependency = subDependencyEdge.Link;
                    if (subDependency == null)
                    {
                        //SKIP null ...
                        continue;
                    }
                    if (bfsVisitedDependencies.Contains(subDependency))
                    {
                        //Already visited ...
                    }
                    else
                    {
                        //This is quite slow ... iterating the queue ... 
                        //Better approach could be importing them to the queue, but during the visisted check they wont be processed anyway ... 
                        //Checking against hash is much faster!!!!
                        /*//New dependency, has to be processed .. or already in the queue?
                        if (bfsToBeAnalysed.Contains(subDependency))
                        {
                            //Is included in to do list ... just reference
                        }
                        else
                        {
                            //Completely new ... will be included in queue
                            bfsToBeAnalysed.Enqueue(subDependency);        //Add to the to process queue

                        }*/
                        bfsToBeAnalysed.Enqueue(subDependency);        //Add to the to process queue
                    }
                }
            }
        }

        /// <summary>
        /// Analyzes one unique dependency of the given data
        /// </summary>
        /// <param name="dependency">Dependency to be analyzed</param>
        private void AnalyzeDependency(DependencyItem dependency)
        {
            //This is the core of the analysis
            //Here will be executed all the analysis 

            foreach (var searchPattern in _searchPatterns)
            {
                if(IsDependencyAlreadyMatched(dependency, searchPattern))
                {
                    //Duplicit dpedendency, already analysed ... no more!
                    continue;
                }

                if (searchPattern.Item1 == "License")
                {
                    if(searchPattern.Item2 == "null" && dependency.License == null)
                    {
                        //Match the null
                        dependency.AnnotationInfo.Add(new AnnotationInfoItem()
                        {
                            AnalyzerName = this.GetType().Name,
                            AnalyzerVersion = Assembly.GetEntryAssembly().GetName().Version,
                            AnalyzerCondition = $"{searchPattern.Item1}={searchPattern.Item2}",
                            InformationType = AnnotationInfoItem.InformationLevel.Info,
                            Text = "Match found!"
                        });
                        MarkTheDependencyVisited(dependency, searchPattern);
                    }
                    else if(searchPattern.Item2 == dependency.License)
                    {
                        //MATCH!!!!!
                        //string matchText = $"Analyzer: {this.GetType().Name} found match using pattern: {searchPattern.Item1}={searchPattern.Item2}";
                        //Console.WriteLine(matchText);
                        dependency.AnnotationInfo.Add(new AnnotationInfoItem() { AnalyzerName = this.GetType().Name, 
                                                                                 AnalyzerVersion = Assembly.GetEntryAssembly().GetName().Version,
                                                                                 AnalyzerCondition = $"{searchPattern.Item1}={searchPattern.Item2}", 
                                                                                 InformationType=AnnotationInfoItem.InformationLevel.Info,
                                                                                 Text = "Match found!"}) ;
                        MarkTheDependencyVisited(dependency, searchPattern);
                    }
                }
            }
        }

        /// <summary>
        /// Adds the analyzed and matched dependency to the hashset and marks it as processed
        /// </summary>
        /// <param name="dependency">Dependency to be marked as processed</param>
        /// <param name="searchPattern">Used search pattern to match</param>
        private void MarkTheDependencyVisited(DependencyItem dependency, Tuple<string, string> searchPattern)
        {
            HashSet<Tuple<string, string>> patternsMarkedSet;

            if (markedNodes.TryGetValue(dependency, out patternsMarkedSet))
            {
                patternsMarkedSet.Add(searchPattern);
            }
            else
            {
                markedNodes.Add(dependency, new HashSet<Tuple<string, string>>());
                markedNodes[dependency].Add(searchPattern);
            }
        }

        /// <summary>
        /// Checks if the dpeendency was already matched using the given pattern
        /// </summary>
        /// <param name="dependency">Dependency to be checked if marked</param>
        /// <param name="searchPattern">Used search pattern to mark</param>
        /// <returns>True, if already matched</returns>
        private bool IsDependencyAlreadyMatched(DependencyItem dependency, Tuple<string, string> searchPattern)
        {
            HashSet<Tuple<string, string>> patternMatchSet;
            if (markedNodes.TryGetValue(dependency, out patternMatchSet))
            {
                if (patternMatchSet.Contains(searchPattern))
                {
                    return true;
                }
            }

            return false;
        }
    }
}
