﻿using Janus.Core.Interfaces;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Janus.Core.Models
{
    /// <summary>
    /// This class represents one project with its dependencies
    /// </summary>
    public class ProjectItem : IComponentItem
    {
        /// <summary>
        /// Constructs and initializes the instance
        /// </summary>
        public ProjectItem()
        {
            Dependencies = new List<DependencyEdge>();
        }

        /// <summary>
        /// Constructs and initializes the instance, sets the given path and obtains project name (Same as project file)
        /// </summary>
        public ProjectItem(string path) : this()
        {
            Path = path;
            if (File.Exists(path))
            {
                Name = System.IO.Path.GetFileNameWithoutExtension(path);
            }
        }

        /// <summary>
        /// Project name
        /// </summary>
        public string Name { get; set; }

        /// <summary>
        /// Project programming language used
        /// </summary>
        public string Language { get; set; }

        /// <summary>
        /// Project programming language version
        /// </summary>
        public string LanguageVersion { get; set; }

        /// <summary>
        /// Project path
        /// </summary>
        public string Path{ get; set; }

        /// <summary>
        /// List of project direct dependencies
        /// Represented using wrapper for the edge
        /// </summary>
        public IList<DependencyEdge> Dependencies { get; set; }

        /// <summary>
        /// Prints the dependency tree in console
        /// </summary>
        /// <param name="indent">Indentation of the node</param>
        /// <param name="last">Marks the item as last, prints end of this branch</param>
        /// <param name="edgeInfo">Additional information about the edge</param>
        public string PrintPretty(string indent, bool last, string edgeInfo = "")
        {
            string dataResult = "";
            //Console.Write(indent);
            dataResult += indent;
            if (last)
            {
                //Console.Write("\\-");
                dataResult += "\\-";
                indent += "  ";
            }
            else
            {
                //Console.Write("|-");
                dataResult += "|-";
                indent += "| ";
            }
            //Console.WriteLine(this);
            dataResult += this.ToString();              //Prints vertex
            if (edgeInfo != null && edgeInfo != "")
            {
                dataResult += $" ({edgeInfo})";         //Adds edge info
            }
            dataResult += "\n";                         //Newline

            for (int i = 0; i < Dependencies.Count; i++)
            {
                dataResult += Dependencies[i].Link.PrintPretty(indent, i == Dependencies.Count - 1, Dependencies[i].RequiredVersionsText);
            }

            return dataResult;
        }

        /// <summary>
        /// Prints object in string format
        /// </summary>
        /// <returns>Textual representation of the ProjectItem</returns>
        public override string ToString()
        {
            return $"{Name} ({Dependencies.Count} direct dependecies)";
        }
    }
}
