﻿using Janus.Core.Models;
using NuGet.Common;
using NuGet.Frameworks;
using NuGet.Packaging.Core;
using NuGet.Protocol;
using NuGet.Protocol.Core.Types;
using NuGet.Versioning;
using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace Janus.Parsers.DotnetCoreNuget.NugetAPI
{
    /// <summary>
    /// This class acts as a wrapper for the nuget dependency requests
    /// </summary>
    class NugetWrapper
    {
        /// <summary>
        /// Default address of the Nuget Server
        /// </summary>
        public static string DefaultNugetAddress { get; } = "https://api.nuget.org/v3/index.json";

        /// <summary>
        /// Private instance of the NugetWrapper used as singleton
        /// </summary>
        private static NugetWrapper _instance = null;

        /// <summary>
        /// Private constructor, to hide from outer code
        /// </summary>
        private NugetWrapper()
        {
        }

        /// <summary>
        /// Returns the instance of the singleton
        /// </summary>
        public static NugetWrapper Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new NugetWrapper();
                }
                return _instance;
            }
        }


        /// <summary>
        /// TESTING ONLY
        /// This method is for testing of the Nuget possibilities, query for package and all its versions
        /// </summary>
        /// <param name="packageName"></param>
        /// <param name="repositoryUrl">Custom repository address</param>
        /// <returns></returns>
        public async Task PrintMetadataAsync(string packageName, string repositoryUrl = null)
        {
            ILogger logger = NullLogger.Instance;
            CancellationToken cancellationToken = CancellationToken.None;

            SourceCacheContext cache = new SourceCacheContext();
            SourceRepository repository = Repository.Factory.GetCoreV3(repositoryUrl ?? DefaultNugetAddress);
            PackageMetadataResource resource = await repository.GetResourceAsync<PackageMetadataResource>();

            IEnumerable<IPackageSearchMetadata> packages = await resource.GetMetadataAsync(
                packageName,
                includePrerelease: true,
                includeUnlisted: false,
                cache,
                logger,
                cancellationToken);


            foreach (IPackageSearchMetadata package in packages)
            {
                Console.WriteLine($"Version: {package.Identity.Version}");
                Console.WriteLine($"Listed: {package.IsListed}");
                Console.WriteLine($"Tags: {package.Tags}");
                Console.WriteLine($"Description: {package.Description}");
                Console.WriteLine($"Authors: {package.Authors}");
                Console.WriteLine($"DependencySets: {package.DependencySets}");
                Console.WriteLine($"DownloadCount: {package.DownloadCount}");
                Console.WriteLine($"IconUrl: {package.IconUrl}");
                Console.WriteLine($"Identity: {package.Identity}");
                Console.WriteLine($"IsListed: {package.IsListed}");
                Console.WriteLine($"LicenseMetadata: {package.LicenseMetadata}");
                Console.WriteLine($"LicenseUrl: {package.LicenseUrl}");
                Console.WriteLine($"Owners: {package.Owners}");
                Console.WriteLine($"PackageDetailsUrl: {package.PackageDetailsUrl}");
                Console.WriteLine($"PrefixReserved: {package.PrefixReserved}");
                Console.WriteLine($"ProjectUrl: {package.ProjectUrl}");
                Console.WriteLine($"Published: {package.Published}");
                Console.WriteLine($"ReadmeUrl: {package.ReadmeUrl}");
                Console.WriteLine($"ReportAbuseUrl: {package.ReportAbuseUrl}");
                Console.WriteLine($"RequireLicenseAcceptance: {package.RequireLicenseAcceptance}");
                Console.WriteLine($"Summary: {package.Summary}");
                Console.WriteLine($"Tags: {package.Tags}");
                Console.WriteLine($"Title: {package.Title}");
                Console.WriteLine($"Vulnerabilities: {package.Vulnerabilities}");

                Console.WriteLine("--------------------------------------------------------------");
            }

        }

        /// <summary>
        /// This method provides the metadata for the package using the Nuget API
        /// </summary>
        /// <param name="packageName">Name of the package</param>
        /// <param name="packageVersion">Version of the package</param>
        /// <param name="repositoryUrl">Custom repository address</param>
        /// <returns></returns>
        public async Task<IPackageSearchMetadata> GetMetadataAsync(string packageName, string packageVersion, string repositoryUrl = null)
        {
            var searchPackage = new PackageIdentity(packageName, NuGetVersion.Parse(packageVersion));
            ILogger logger = NullLogger.Instance;
            CancellationToken cancellationToken = CancellationToken.None;

            SourceCacheContext cache = new SourceCacheContext();
            SourceRepository repository = Repository.Factory.GetCoreV3(repositoryUrl ?? DefaultNugetAddress);
            PackageMetadataResource resource = await repository.GetResourceAsync<PackageMetadataResource>();

            var package = await resource.GetMetadataAsync(
                searchPackage,
                cache,
                logger,
                cancellationToken
                );
            /*
            Console.WriteLine($"Version: {package.Identity.Version}");
            Console.WriteLine($"Listed: {package.IsListed}");
            Console.WriteLine($"Tags: {package.Tags}");
            Console.WriteLine($"Description: {package.Description}");
            Console.WriteLine($"Authors: {package.Authors}");
            Console.WriteLine($"DependencySets: {package.DependencySets}");
            Console.WriteLine($"DownloadCount: {package.DownloadCount}");
            Console.WriteLine($"IconUrl: {package.IconUrl}");
            Console.WriteLine($"Identity: {package.Identity}");
            Console.WriteLine($"IsListed: {package.IsListed}");
            Console.WriteLine($"LicenseMetadata: {package.LicenseMetadata}");
            Console.WriteLine($"LicenseUrl: {package.LicenseUrl}");
            Console.WriteLine($"Owners: {package.Owners}");
            Console.WriteLine($"PackageDetailsUrl: {package.PackageDetailsUrl}");
            Console.WriteLine($"PrefixReserved: {package.PrefixReserved}");
            Console.WriteLine($"ProjectUrl: {package.ProjectUrl}");
            Console.WriteLine($"Published: {package.Published}");
            Console.WriteLine($"ReadmeUrl: {package.ReadmeUrl}");
            Console.WriteLine($"ReportAbuseUrl: {package.ReportAbuseUrl}");
            Console.WriteLine($"RequireLicenseAcceptance: {package.RequireLicenseAcceptance}");
            Console.WriteLine($"Summary: {package.Summary}");
            Console.WriteLine($"Tags: {package.Tags}");
            Console.WriteLine($"Title: {package.Title}");
            Console.WriteLine($"Vulnerabilities: {package.Vulnerabilities}");

            Console.WriteLine("--------------------------------------------------------------");*/

            return package;
        }

        /// <summary>
        /// Fills the DependencyItem properties with the data obtainer using the method parameters and data from remote repository. 
        /// Internally calls <see cref="GetMetadataAsync"/> to obtain the remote data.
        /// </summary>
        /// <param name="packageName">Name of the package, ID</param>
        /// <param name="packageVersion">Version of the package</param>
        /// <param name="repositoryUrl">Custom repository address</param>
        /// <returns>DependencyItem filled with data</returns>
        public async Task<DependencyItem> FillMetadataAsync(string packageName, string packageVersion, string repositoryUrl = null)
        {
            DependencyItem depItem = new DependencyItem();
            depItem.Id = packageName;
            depItem.Name = packageName;
            depItem.Version = packageVersion;
            depItem.RegistryType = "Nuget";

            var packageMetadata = await GetMetadataAsync(depItem.Id, depItem.Version, repositoryUrl);

            if (!string.IsNullOrEmpty(packageMetadata.Authors))
            {
                depItem.Authors.Add(packageMetadata.Authors);
            }
            depItem.License = packageMetadata.LicenseMetadata?.License;
            depItem.LicenseLink = packageMetadata.LicenseMetadata?.LicenseUrl?.AbsoluteUri ?? packageMetadata.LicenseUrl?.AbsoluteUri; //Save licence from metadata, if not possible try to save the unspecified license link
            depItem.Summary = packageMetadata.Summary;
            depItem.Description = packageMetadata.Description;
            depItem.Link = packageMetadata.PackageDetailsUrl.ToString();

            return depItem;
        }

        /// <summary>
        /// This method performs the search on the remote Nuget server
        /// This search will try to find subdependencies of the given dependency
        /// </summary>
        /// <param name="packageId">Id of the dependency to be analyzed</param>
        /// <param name="packageVersion">Version of the dependency to be analyzed</param>
        /// <param name="framework">Framework used in the project, to find the right dependency match</param>
        /// <param name="repositoryUrl">Custom repository address</param>
        /// <returns></returns>
        public async Task<SourcePackageDependencyInfo> GetDependencyDataAsync(string packageId, string packageVersion, string framework, string repositoryUrl = null)
        {
            var package = new PackageIdentity(packageId, NuGetVersion.Parse(packageVersion));
            SourceRepository repository = Repository.Factory.GetCoreV3(repositoryUrl ?? DefaultNugetAddress); //TODO: Multiple repositories? Is it even possible?
            var nuGetFramework = NuGetFramework.ParseFolder(framework);
            var logger = NullLogger.Instance;

            using (var cacheContext = new SourceCacheContext())
            {
                var dependencyInfoResource = await repository.GetResourceAsync<DependencyInfoResource>();
                var dependencyInfo = await dependencyInfoResource.ResolvePackage(
                    package, nuGetFramework, cacheContext, logger, CancellationToken.None);
                    
                return dependencyInfo;
            }
        }
    }
}
