﻿using Janus.Core.Models;
using Janus.Core.Models.Results;
using Janus.Parsers.PythonPyPi.PyPIAPI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Janus.Parsers.PythonPyPi.Parser
{
    /// <summary>
    /// This class helps with the parsing of the PyPI based projects
    /// </summary>
    public class PyPIParser
    {

        /// <summary>
        /// Instance of the pypi wrapper
        /// </summary>
        private static PyPIAPI.PyPIWrapper _pypiWrapper = PyPIAPI.PyPIWrapper.Instance;

        /// <summary>
        /// Cache of the visited dependencies to be linked. This cache reduces the API calls to absolute minimum
        /// Key:   Tuple<IdPackage, VersionPackage>
        /// Value: objectReference
        /// </summary>
        private static Dictionary<Tuple<string, string>, DependencyItem> _cachedDependencies = new Dictionary<Tuple<string, string>, DependencyItem>();

        /// <summary>
        /// Private constructor, this should not be instanced
        /// </summary>
        private PyPIParser()
        {

        }

        /// <summary>
        /// This method parses the dependencies of the given package indetifiers acting as roots
        /// Retuns the same roots, but parsed as the DependencyItems
        /// </summary>
        /// <param name="packages">Packages to be used as roots</param>
        /// <returns></returns>
        public static GenericResult<List<DependencyItem>> ParseDependencies(List<Tuple<string, string>> packages)
        {
            GenericResult<List<DependencyItem>> result = new GenericResult<List<DependencyItem>>();

            GenericResult<List<PyPIDepTreeItem>> treePipgripResult = _pypiWrapper.GetDepTreeForPackages(packages);

            if (!treePipgripResult.OperationSuccessful)
            {
                result.OperationSuccessful = false;
                result.Errors = treePipgripResult.Errors; //Forward errors
                return result;
            }

            //Pipgrip done and OK

            List<DependencyItem> dependencies = new List<DependencyItem>(); //List for the roots

            foreach (var pipgripRoot in treePipgripResult.Result)
            {
                //BFS For each root node
                DependencyItem item = ProcessDependencyItem(pipgripRoot);
                dependencies.Add(item);
            }

            result.OperationSuccessful = true;
            result.Result = dependencies;

            return result;
        }

        /// <summary>
        /// Using the given root item, parses the graph to the dependecyItem structure
        /// Also fills all the metadata that are available from the remotes
        /// Uses the resolve cache, if item was analysed it is only referenced fro mcache, no further annalysis
        /// </summary>
        /// <param name="pipgripRoot">Root of the graph</param>
        /// <returns></returns>
        static DependencyItem ProcessDependencyItem(PyPIDepTreeItem pipgripRoot)
        {
            DependencyItem dummyNode = new DependencyItem();

            //Job tuple: To be visited with link to the parent (Null for the root)
            Queue<Tuple<PyPIDepTreeItem, DependencyItem>> toBeVisited = new Queue<Tuple<PyPIDepTreeItem, DependencyItem>>();

            toBeVisited.Enqueue(Tuple.Create(pipgripRoot, dummyNode)); //First item to be processed

            while (toBeVisited.Any())
            {
                Tuple<PyPIDepTreeItem, DependencyItem> currentPair = toBeVisited.Dequeue();
                PyPIDepTreeItem current = currentPair.Item1;
                DependencyItem currentParent = currentPair.Item2;

                var index = new Tuple<string, string>(current.name, current.version);

                //First check if the dependency is not in cache
                if (_cachedDependencies.ContainsKey(index))
                {
                    //Already in cache!!! Just reference that and continue
                    AddPyPINodeToDependencyItem(current, currentParent, _cachedDependencies[index]); //Already in the cache, just reference that
                    continue;
                }

                //Not in the cache, has to be added and childs has to be examined

                DependencyItem parsedNode = AddPyPINodeToDependencyItem(current, currentParent);

                foreach (var child in current.dependencies)
                {
                    toBeVisited.Enqueue(Tuple.Create(child, parsedNode)); //add job item and its parent
                }
            }

            return dummyNode.Dependencies.First().Link;
        }

        /// <summary>
        /// Converts the node to the DependencyItem type and adds it to the parent with the coresponding edge limits
        /// If node is already parsed and is just needed to stitch to the parent, it can be provided using parsedNode parameter
        /// </summary>
        /// <param name="node">Node to be converted and added to the parent</param>
        /// <param name="parent">Parent node</param>
        /// <param name="parsedNode">Already parsed node provider (Skips metadata download)</param>
        /// <returns></returns>
        public static DependencyItem AddPyPINodeToDependencyItem(PyPIDepTreeItem node, DependencyItem parent, DependencyItem parsedNode = null)
        {
            DependencyItem workDependency = null;

            if(parsedNode != null)
            {
                workDependency = parsedNode;
            }
            else
            {
                workDependency = new DependencyItem();
                //TODO Solve obtain fail
                GenericResult<PyPIMetadataRoot> metaResult = _pypiWrapper.GetPackageMetadata(node.name, node.version);
                PyPIMetadataRoot metadata = metaResult.Result;

                workDependency.Id = metadata.info.name;
                workDependency.Name = metadata.info.name;
                workDependency.Version = metadata.info.version;
                workDependency.RegistryType = "PyPI";
                workDependency.Description = metadata.info.description;
                workDependency.Summary = metadata.info.summary;
                if (metadata.info.author != null)
                {
                    workDependency.Authors.Add(metadata.info.author);
                }
                workDependency.License = string.IsNullOrEmpty(metadata.info.license) ? null : metadata.info.license;
                workDependency.Link = metadata.info.release_url;
                workDependency.PackageDownloadLink = metadata.urls.Where(x => x.packagetype == "source").FirstOrDefault()?.url;

                //Cache the result
                var index = new Tuple<string, string>(workDependency.Id, workDependency.Version);
                _cachedDependencies[index] = workDependency;
            }

            //We have dependency now just add edge
            DependencyEdge edge = new DependencyEdge();
            edge.Parent = parent;
            edge.Link = workDependency;

            //Construct limits
            bool firstLimit = true;
            foreach (var item in node.pip_string.Split(","))
            {
                string workItem = item;

                if (firstLimit)
                {
                    firstLimit = false;
                    workItem = workItem.Replace(node.name, ""); //Replace occurences
                }

                VersionLimit newLimit = new VersionLimit();

                if (workItem.StartsWith(">="))
                {
                    workItem = workItem.Remove(0, 2);
                    newLimit.LimitSymbol = VersionLimit.OperatorType.GreaterOrEqual;
                }
                else if (workItem.StartsWith("<="))
                {
                    workItem = workItem.Remove(0, 2);
                    newLimit.LimitSymbol = VersionLimit.OperatorType.LowerOrEqual;
                }
                else if (workItem.StartsWith("=="))
                {
                    workItem = workItem.Remove(0, 2);
                    newLimit.LimitSymbol = VersionLimit.OperatorType.Equal;
                }
                else if (workItem.StartsWith("~="))
                {
                    workItem = workItem.Remove(0, 2);
                    newLimit.LimitSymbol = VersionLimit.OperatorType.Compatible;
                }
                else if(workItem.StartsWith("!="))
                {
                    workItem = workItem.Remove(0, 2);
                    newLimit.LimitSymbol = VersionLimit.OperatorType.NotEqual;
                }
                else if (workItem.StartsWith(">"))
                {
                    workItem = workItem.Remove(0, 1);
                    newLimit.LimitSymbol = VersionLimit.OperatorType.Greater;
                }
                else if (workItem.StartsWith("<"))
                {
                    workItem = workItem.Remove(0, 1);
                    newLimit.LimitSymbol = VersionLimit.OperatorType.Lower;
                }
                else 
                {
                    //Dunno
                    continue;
                }

                newLimit.RequiredVersion = workItem;
                edge.VersionLimits.Add(newLimit);
            }

            parent.Dependencies.Add(edge);

            return workDependency;
        }
    }
}
