﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using Janus.Core.Models;
using Janus.Core.Models.Results;

namespace Janus.Parsers.PythonPyPi.PyPIAPI
{
    /// <summary>
    /// This class acts as a wrapper for the nuget dependency requests
    /// </summary>
    class PyPIWrapper
    {
        /// <summary>
        /// Default address of the PyPI Server
        /// </summary>
        public static string DefaultPyPIAddress { get; } = "https://pypi.org/simple";

        /// <summary>
        /// Private instance of the PyPIWrapper used as singleton
        /// </summary>
        private static PyPIWrapper _instance = null;

        /// <summary>
        /// Private constructor, to hide from outer code
        /// </summary>
        private PyPIWrapper()
        {
        }

        /// <summary>
        /// Returns the instance of the singleton
        /// </summary>
        public static PyPIWrapper Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new PyPIWrapper();
                }
                return _instance;
            }
        }

        /// <summary>
        /// Obtains dependency tree for the given packages, action as roots
        /// </summary>
        /// <param name="packages">Packages to be used as roots</param>
        /// <returns></returns>
        public GenericResult<List<PyPIDepTreeItem>> GetDepTreeForPackages(List<Tuple<string,string>> packages)
        {
            GenericResult<List<PyPIDepTreeItem>> result = new GenericResult<List<PyPIDepTreeItem>>();

            string packagesDefinition = "";

            foreach (var item in packages)
            {
                packagesDefinition += $" {item.Item1}=={item.Item2}";
            }

            ProcessStartInfo procStartInfo = new ProcessStartInfo("pipgrip");

            procStartInfo.RedirectStandardOutput = true;
            procStartInfo.RedirectStandardError = true;
            procStartInfo.UseShellExecute = false;
            procStartInfo.CreateNoWindow = true;
            procStartInfo.Arguments = Encoding.Default.GetString(Encoding.UTF8.GetBytes($"--tree --json{packagesDefinition}"));

            // wrap IDisposable into using (in order to release hProcess) 
            using (Process process = new Process())
            {
                string resultProcess = "";
                string resultErrProcess = "";
                int exitCode = -1;
                process.StartInfo = procStartInfo;
                try
                {
                    process.Start();

                    // Add this: wait until process does its work
                    process.WaitForExit();

                    exitCode = process.ExitCode;

                    // and only then read the result
                    resultProcess = process.StandardOutput.ReadToEnd();
                    resultErrProcess = process.StandardError.ReadToEnd();
                }
                catch (Exception e)
                {
                    //Pipgrip not installed or not in path
                    result.OperationSuccessful = false;
                    result.Errors.Add("Error: Cant find pipgrip, probably not installed or not in the path!");
                    result.Errors.Add(e.Message);
                    return result;
                }

                if(exitCode != 0)
                {
                    //Exited with non zero ...
                    result.OperationSuccessful = false;
                    result.Errors.Add("Error: pipgrip ended with non zero exit code!");
                    result.Errors.Add(resultErrProcess);
                    return result;
                }

                try
                {
                    List<PyPIDepTreeItem> deps = JsonConvert.DeserializeObject<List<PyPIDepTreeItem>>(resultProcess);
                    result.OperationSuccessful = true;
                    result.Result = deps;

                }
                catch (Exception e)
                {
                    //Cannot deserialize
                    result.OperationSuccessful = false;
                    result.Errors.Add($"Error: cannot deserialize result of pipgrip! (Result: {resultErrProcess})");
                    result.Errors.Add(e.Message);
                    return result;
                }
            }

            return result;
        }

        /// <summary>
        /// Obtains metadata for the specific package
        /// </summary>
        /// <param name="id">Identifier name of the package</param>
        /// <param name="version">Version of the package</param>
        /// <returns></returns>
        public GenericResult<PyPIMetadataRoot> GetPackageMetadata(string id, string version)
        {
            GenericResult<PyPIMetadataRoot> result = new GenericResult<PyPIMetadataRoot>();
            string url = $"https://pypi.org/pypi/{id}/{version}/json";
            string html = string.Empty;

            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
            request.AutomaticDecompression = DecompressionMethods.GZip | DecompressionMethods.Deflate; //In case of server compression

            using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
            using (Stream stream = response.GetResponseStream())
            using (StreamReader reader = new StreamReader(stream))
            {
                html = reader.ReadToEnd();
            }

            try
            {
                PyPIMetadataRoot deps = JsonConvert.DeserializeObject<PyPIMetadataRoot>(html);
                result.Result = deps;
                result.OperationSuccessful = true;
            }
            catch (Exception e)
            {
                result.OperationSuccessful = false;
                result.Errors.Add(e.Message);
            }

            return result;
        }
    }
}
